#region usings
using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Linq;

using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.Graph;
using VVVV.Core;
using VVVV.Core.View;
using VVVV.Core.Logging;
using VVVV.Utils.Streams;
using VVVV.Utils.Linq;
using VVVV.PatchMessage;
using VVVV.Snap;

#endregion usings

namespace VVVV.Nodes
{
	#region PluginInfo
	
	[PluginInfo(Name = "PatchManager", Category = "VVVV", Help = "Provides Snap and Grid for Patches and Nodes", Tags = "" , AutoEvaluate = true)]
	#endregion PluginInfo
	public class PatchManagerNode:   IDisposable ,IPartImportsSatisfiedNotification, IPluginEvaluate
	{
		
		#region field declaration
		[Import()]
		public ILogger FLogger;
		
		[Import()]
		private IHDEHost FHDEHost;
		private IPluginHost2 FPluginHost;
			
		// SNAP
		SnapWindows 			Window;
		SnapPatchsAndLinks		Patch;
		string					vvvvExePath;
		
		// HOOK
		private const int WH_KEYBOARD_LL = 13;
		private const int WM_KEYDOWN = 0x0100;
		private const int WM_KEYUP = 0x0101;
		private const int WM_SYSKEYDOWN = 0x0104;
		private const int WM_SYSKEYUP 	= 0x0105;
		private IntPtr windowHandle;
		private LowLevelKeyboardProc _proc;
		private IntPtr _hookID = IntPtr.Zero;		
		private bool disableHook = false;	
		#endregion field declaration
		
		#region  pins
		[Input("Enable", IsSingle = true, Order = 0)]
		IDiffSpread<bool> FEnable;
		
		[Input("Inspector ID", IsSingle = true, Order = 1)]
		IDiffSpread<int> FInspektor;
		
		[Input("NodeBrowser ID", IsSingle = true, Order = 2)]
		IDiffSpread<int> FNodeBrowser;
		
		[Input("Finder ID", IsSingle = true, Order = 3)]
		IDiffSpread<int> FFinder;
		
		[Input("Projektexplorer ID", IsSingle = true, Order = 4)]
		IDiffSpread<int> FProjektexplorer;
		
		[Input("Max Pixel to Snap", DefaultValue = 100, IsSingle = true, Order = 5)]
		public IDiffSpread<int> FDeltaMax;
		
		[Output( "ShortCutNumber", Order = 0)]
		ISpread<int> FShortCutNumber;
		
		[Output("IgnoreKeys", Order = 1)]
		public ISpread<int> FIgnoreKeys;
		
		[Output("PressedKeys", Order = 2)]
		public ISpread<int> FPressedKeys;
		
		[Output("XML_OUT", Order = 8)]
		public ISpread<string> FOutXML;
		
		[Output("horizontal SnapPoints", Order = 9)]
		public ISpread<int> FOutHorizontalSnaps;
		
		[Output("vertical SnapPoints", Order = 10)]
		public ISpread<int> FOutVerticalSnaps;		
		
		[Output("DEBUG", Order = 12)]
		public ISpread<string> FOutDebug;
		
		#endregion pins
		
		#region constructor/destructor
		[ImportingConstructor]
		public PatchManagerNode(IHDEHost host, IPluginHost2 pluginHost)
		{
			FHDEHost	= host;
			Patch		= new SnapPatchsAndLinks(host);
			Window		= new SnapWindows(host, this);
			FPluginHost = pluginHost;
			vvvvExePath = FHDEHost.ExePath + @"\" + "vvvv.exe";
		}
		
		public void OnImportsSatisfied() {
			FShortCutNumber[0]	= -1;
			FIgnoreKeys.SliceCount = 0;
			FPressedKeys.SliceCount = 0;				
		}
		
		
		public void Dispose()
		{
			UnsetHook();
		}
		#endregion constructor/destructor
		
		#region Evaluate
		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
			// SetHOOK
			if(FEnable.IsChanged)
			{
				if(_hookID != IntPtr.Zero)
				UnsetHook();
				if(FEnable[0])
				SetHook();
			}
			
			
			// Window changed
			if (GetModuleHandle(GetWindowModuleFileName(GetForegroundWindow())) != windowHandle)
			{
				
				disableHook = true;
			} else {
				disableHook = false;
			}
			
			if(FInspektor.IsChanged)
			{
				string NodeIDPath	= FPluginHost.GetNodePath(false);
				string PatchIDPath	= NodeIDPath.Substring(0, NodeIDPath.LastIndexOf('/')+1);
				Window.inspektor = FHDEHost.GetNodeFromPath(PatchIDPath + FInspektor[0].ToString());
			}
			
			if(FNodeBrowser.IsChanged)
			{
				string NodeIDPath	= FPluginHost.GetNodePath(false);
				string PatchIDPath	= NodeIDPath.Substring(0, NodeIDPath.LastIndexOf('/')+1);
				Window.nodeBrowser = FHDEHost.GetNodeFromPath(PatchIDPath + FNodeBrowser[0].ToString());
			}
			
			if(FFinder.IsChanged)
			{
				string NodeIDPath	= FPluginHost.GetNodePath(false);
				string PatchIDPath	= NodeIDPath.Substring(0, NodeIDPath.LastIndexOf('/')+1);
				Window.finder = FHDEHost.GetNodeFromPath(PatchIDPath + FFinder[0].ToString());
			}
			
			if(FProjektexplorer.IsChanged)
			{
				string NodeIDPath	= FPluginHost.GetNodePath(false);
				string PatchIDPath	= NodeIDPath.Substring(0, NodeIDPath.LastIndexOf('/')+1);
				Window.projektexplorer = FHDEHost.GetNodeFromPath(PatchIDPath + FProjektexplorer[0].ToString());
			}/*
*/

			
			if (FShortCutNumber[0]  >= 0 && FIgnoreKeys.IsChanged)
			{
		
				switch (FShortCutNumber[0])
				{
					// WIN
					case  0: 									break;
					
					// WIN + SPACE
					case  1: Window.SnapActiveWindow(FDeltaMax[0]);	break;
					
					// WIN + SHIFT + SPACE
					case  2: Window.SnapActiveWindow(FDeltaMax[0]);	break;
					
					// WIN + LEFT
					case  3: Window.goLeft();					break;
					// WIN + SHIFT+ LEFT
					case  4: Window.crowLeft();					break;
					
					// WIN + RIGHT
					case  5: Window.goRight();					break;
					// WIN + SHIFT+ RIGHT
					case  6: Window.crowRight();				break;
					
					// WIN + UP
					case  7: Window.goUp();						break;
					// WIN + SHIFT+ UP
					case  8: Window.crowUp();					break;
					
					// WIN + DOWN
					case  9: Window.goDown();					break;
					// WIN + SHIFT+ DOWN
					case 10: Window.crowDown();					break;
					
					// WIN + [Y]
					case 11: Patch.SnapLink(true);				break;
					// WIN + SHIFT+ [Y]
					case 12: Patch.SnapLink(false);				break;
					
					// WIN + [<]
					case 13: Patch.SnapNode(15,150,12,240);		break;
					// WIN + SHIFT+ [<]
					case 14: Patch.SnapNode(150,150,240,240);	break;
					
					case 51: Window.ShowInspektor();			break; // INSPECTOR
					case 52: Window.ShowNodeBrowser();			break; // NODEBROWSER
					case 53: Window.ShowFinder();				break; // FINDER
					case 54: Window.ShowProjektexplorer();		break; // PROJEKTEXPLORER
					case 55: Window.Hide();						break; // HIDE
					
					// DEFAULT
					default: 													break;
				}
				if (FShortCutNumber[0] >= 11 && FShortCutNumber[0] <= 14)
				{
					FOutXML.SliceCount = 0;
					foreach(string message in Patch.FOutXML)
					{
						FOutXML.Add(message);
					}
				} else {
					if (Window.FOutXML.Count > 0)
					{
						FOutXML.SliceCount = 0;
						for(int i = 0; i < Window.FOutXML.Count ;i++)
						{
							FOutXML.Add(Window.FOutFile[i]+ "\r\n" +Window.FOutXML[i].ToString());
						}
						Window.FOutFile.Clear();
						Window.FOutXML.Clear();
					}
					FOutHorizontalSnaps.SliceCount = 0;
					foreach(int message in Window.FOutHorizontalSnaps)
					{
						FOutHorizontalSnaps.Add(message);
					}
					FOutVerticalSnaps.SliceCount = 0;
					foreach(int message in Window.FOutVerticalSnaps)
					{
						FOutVerticalSnaps.Add(message);
					}
				}
			}
		}
		#endregion Evaluate
		
		#region KeyHook
		
		public void SetHook() {
			if((_proc) == null)_proc = HookCallback;
			if(_hookID != IntPtr.Zero)
			UnsetHook();
			
			_hookID = SetHook(_proc);
		}
		
		public void UnsetHook() {
			UnhookWindowsHookEx(_hookID);
			_hookID = IntPtr.Zero;
		}
		
		private IntPtr SetHook(LowLevelKeyboardProc proc)
		{
			using (Process curProcess = Process.GetCurrentProcess())
			using (ProcessModule curModule = curProcess.MainModule)
			{
				windowHandle = GetModuleHandle(curModule.ModuleName);
				//FLogger.Log(LogType.Debug, "Hook hooked.");
				return SetWindowsHookEx(
				WH_KEYBOARD_LL,
				proc,
				windowHandle,
				0);
			}
		}
		
		private delegate IntPtr LowLevelKeyboardProc(int nCode, IntPtr wParam, IntPtr lParam);
		
		private void AddFIgnoreKeys(IntPtr wParam,int vkCode)
		{
			if ((wParam == (IntPtr)WM_KEYDOWN || wParam == (IntPtr)WM_SYSKEYDOWN) && FIgnoreKeys.IndexOf(vkCode) < 0)
			{
				FIgnoreKeys.Add(vkCode);
				GetFShortCutNumber();
			}
		}
		
		private void RemoveFIgnoreKeys(IntPtr wParam,int vkCode)
		{
			if (wParam == (IntPtr)WM_KEYUP || wParam == (IntPtr)WM_SYSKEYUP)
			{
				FIgnoreKeys.Remove(vkCode);
				if (vkCode == 91)
				{
					FIgnoreKeys.SliceCount = 0; 
				}
				GetFShortCutNumber();
			}
		}
		
		private void SetFPressedKeys(IntPtr wParam,int vkCode)
		{
			if (wParam == (IntPtr)WM_KEYUP || wParam == (IntPtr)WM_SYSKEYUP)
			{
				FPressedKeys.Remove(vkCode);
			}
			else if ((wParam == (IntPtr)WM_KEYDOWN || wParam == (IntPtr)WM_SYSKEYDOWN) && FPressedKeys.IndexOf(vkCode) < 0 && FIgnoreKeys.IndexOf(vkCode) < 0)
			{
				FPressedKeys.Add(vkCode);
			}
		}
		
		private IntPtr HookCallback(int nCode, IntPtr wParam, IntPtr lParam)
		{
			if (disableHook)
			{
				return CallNextHookEx(_hookID, nCode, wParam, lParam);
			}
			
			
			if ( nCode >= 0)
			{
				bool breakIT = false;
				int vkCode = Marshal.ReadInt32(lParam);
				// exchange special keys
				if(vkCode == 162 || vkCode == 163) vkCode = 17; 	// left ctrl + right ctrl
				if(vkCode == 160 || vkCode == 161) vkCode = 16; 	// left shift + right shift
				if(vkCode == 164 || vkCode == 165) vkCode = 18;		// alt
				
				if ((wParam == (IntPtr)WM_KEYUP || wParam == (IntPtr)WM_SYSKEYUP)&&((vkCode == 91)||(vkCode == 18)||(vkCode == 17)))
				{
					// UNHOOK if WIN, ALT, Ctrl released
					FPressedKeys.SliceCount = 0;
					FIgnoreKeys.SliceCount = 0;
					return CallNextHookEx(_hookID, nCode, wParam, lParam);
				}
				
				if (FIgnoreKeys.SliceCount == 0)
				{
					if (vkCode == 91)
					{
						AddFIgnoreKeys(wParam,vkCode);
						breakIT = true;
					}
					/*else if	((vkCode == 87) && (FPressedKeys.SliceCount == 1 && FPressedKeys[0] == 17) && (Window.FActiveWindow.Handle == Window.inspektor.Window.Handle || Window.FActiveWindow.Handle == Window.finder.Window.Handle || Window.FActiveWindow.Handle == Window.projektexplorer.Window.Handle || Window.FActiveWindow.Handle == Window.nodeBrowser.Window.Handle))
					{
						AddFIgnoreKeys(wParam,vkCode);
						breakIT = true;//return (IntPtr)1;
					}*/
					else if	((vkCode == 73||vkCode == 78||vkCode == 70||vkCode == 74) && (FPressedKeys.SliceCount == 1 && FPressedKeys[0] == 17))
					{
						AddFIgnoreKeys(wParam,vkCode);
						breakIT = true;//return (IntPtr)1;
					}
					else if	(vkCode == 51 && (FPressedKeys.SliceCount == 1 && FPressedKeys[0] == 18))
					{
						AddFIgnoreKeys(wParam,vkCode);
						breakIT = true;//return (IntPtr)1;
					}
					
					SetFPressedKeys(wParam,vkCode);
				}
				else if (FIgnoreKeys.SliceCount == 1)
				{
					if ((FIgnoreKeys[0]==91)&&(vkCode == 16||vkCode == 9||vkCode == 32||vkCode == 37||vkCode == 38||vkCode == 39||vkCode == 40||vkCode == 89||vkCode == 226))
					{
						AddFIgnoreKeys(wParam,vkCode);
						breakIT = true;//return (IntPtr)1;
					}
					if ((FIgnoreKeys[0]==91)&&(vkCode == 91))
					{
						breakIT = true;//return (IntPtr)1;
					}
					RemoveFIgnoreKeys(wParam,vkCode);
					SetFPressedKeys(wParam,vkCode);
				}
				else if (FIgnoreKeys.SliceCount == 2)
				{
					if ((FIgnoreKeys[0]==91&&FIgnoreKeys[1]==16)&&(vkCode == 9||vkCode == 32||vkCode == 37||vkCode == 38||vkCode == 39||vkCode == 40||vkCode == 89||vkCode == 226))
					{
						AddFIgnoreKeys(wParam,vkCode);
						breakIT = true;//return (IntPtr)1;
					}
					RemoveFIgnoreKeys(wParam,vkCode);
					SetFPressedKeys(wParam,vkCode);
					if (FIgnoreKeys.IndexOf(vkCode) == 1) breakIT = true;//return (IntPtr)1;
					
				}
				else if (FIgnoreKeys.SliceCount == 3)
				{
					RemoveFIgnoreKeys(wParam,vkCode);
					SetFPressedKeys(wParam,vkCode);
					if (FIgnoreKeys.IndexOf(vkCode) == 2) breakIT = true;//return (IntPtr)1;
				}
				if (breakIT) return (IntPtr)1;
			}
			return CallNextHookEx(_hookID, nCode, wParam, lParam);
			
		}
		
		private void GetFShortCutNumber()
		{
			//FShortCutNumber	= new List<int>(1);
			
			if (FIgnoreKeys.SliceCount == 0)
			{
				FShortCutNumber[0] = -1;
			}
			else if (FIgnoreKeys.SliceCount == 1)
			{
				switch (FIgnoreKeys[0])
				{
					case  91:FShortCutNumber[0]= 0; break; // WIN
					case  73:FShortCutNumber[0]=51; break; // INSPECTOR
					case  78:FShortCutNumber[0]=52; break; // NODEBROWSER
					case  70:FShortCutNumber[0]=53; break; // FINDER
					case  74:FShortCutNumber[0]=54; break; // PROJEKTEXPLORER
					case  51:FShortCutNumber[0]=55; break; // HIDE
					default:break;
				}
			}
			else if (FIgnoreKeys.SliceCount == 2 && FIgnoreKeys[0]==91)
			{
				switch (FIgnoreKeys[1])
				{
					case   9:FShortCutNumber[0]=60; break; // WIN + TAB
					case  16:FShortCutNumber[0]= 0; break; // WIN + SHIFT
					case  32:FShortCutNumber[0]= 1; break; // WIN + SPACE
					case  37:FShortCutNumber[0]= 3; break; // WIN + LEFT
					case  39:FShortCutNumber[0]= 5; break; // WIN + RIGHT
					case  38:FShortCutNumber[0]= 7; break; // WIN + UP
					case  40:FShortCutNumber[0]= 9; break; // WIN + DOWN
					case  89:FShortCutNumber[0]=11; break; // WIN + Y
					case 226:FShortCutNumber[0]=13; break; // WIN + <
					default:break;
				}
			}
			else if (FIgnoreKeys.SliceCount == 3 && FIgnoreKeys[0]==91 && FIgnoreKeys[1]==16)
			{
				switch (FIgnoreKeys[2])
				{
					case   9:FShortCutNumber[0]=61; break; // WIN + SHIFT + TAB
					case  32:FShortCutNumber[0]= 2; break; // WIN + SHIFT + SPACE
					case  37:FShortCutNumber[0]= 4; break; // WIN + SHIFT + LEFT
					case  39:FShortCutNumber[0]= 6; break; // WIN + SHIFT + RIGHT
					case  38:FShortCutNumber[0]= 8; break; // WIN + SHIFT + UP
					case  40:FShortCutNumber[0]=10; break; // WIN + SHIFT + DOWN
					case  89:FShortCutNumber[0]=12; break; // WIN + SHIFT + Y
					case 226:FShortCutNumber[0]=14; break; // WIN + SHIFT + <
					default:break;
				}
			} else {
				FShortCutNumber[0] = -1;
			}
			//Execute();
		}
		
		#endregion KeyHook
		
		#region DllImport
		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr SetWindowsHookEx(int idHook,
		LowLevelKeyboardProc lpfn, IntPtr hMod, uint dwThreadId);
		
		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool UnhookWindowsHookEx(IntPtr hhk);
		
		[DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr CallNextHookEx(IntPtr hhk, int nCode,
		IntPtr wParam, IntPtr lParam);
		
		[DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern IntPtr GetModuleHandle(string lpModuleName);
		
		[DllImport("user32.dll")]
		private static extern IntPtr GetForegroundWindow();
		
		[DllImport("user32.dll")]
		private static extern bool SetForegroundWindow(IntPtr hWnd);
		
		[DllImport("user32.dll")]
		private static extern bool GetParent(IntPtr hWnd);
		
		[DllImport("user32.dll")]
		private static extern int GetWindowModuleFileName(IntPtr hWnd, StringBuilder text, int count);
		
		[DllImport("user32.dll", SetLastError = true)]
        private static extern bool MoveWindow(IntPtr hWnd, int X, int Y, int Width, int Height, bool Repaint);
		
		private string GetWindowModuleFileName(IntPtr handle)
		{
			const int nChars = 256;
			StringBuilder Buff = new StringBuilder(nChars);
			
			if (GetWindowModuleFileName(handle, Buff, nChars) > 0)
			{
				return Buff.ToString();
			}
			return "NONE";
		}
		
		[DllImport("user32.dll")]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool ShowWindow(IntPtr hWnd, uint nCmdShow);
		
		#endregion DllImport
	}
}