#region usings
using System;
using System.Drawing;
using System.Linq;
using System.Xml;
using VVVV.PluginInterfaces.V2;
#endregion usings

namespace VVVV.PatchMessage
{
	#region VVVV.xmlMessage
	public class PatchMessageExt
	{
		private XmlDocument FPatch;
		public XmlElement XML
		{
			get {return FPatch.DocumentElement;}
		}
		
		public PatchMessageExt()
		{
			FPatch = new XmlDocument();
			var patch = FPatch.CreateElement("PATCH");
			FPatch.AppendChild(patch);
		}
		
		public PatchMessageExt(string fileName)
		{
			FPatch = new XmlDocument();
			var patch = FPatch.CreateElement("PATCH");
			FPatch.AppendChild(patch);
			Filename = fileName;
		}
		
		public string Filename
		{
			get {return FPatch.DocumentElement.GetAttribute("saveme");}
			set
			{
				if (!string.IsNullOrEmpty(value))
				FPatch.DocumentElement.SetAttribute("saveme", value);
			}
		}
		
		public NodeMessageExt AddNode(int id)
		{
			var node = new NodeMessageExt(FPatch, id);
			FPatch.DocumentElement.AppendChild(node);
			return node;
		}
		
		public NodeMessageExt AddNode(string systemName)
		{
			var node = new NodeMessageExt(FPatch, -1);
			node.SystemName = systemName;
			FPatch.DocumentElement.AppendChild(node);
			return node;
		}
		
		public LinkMessageExt AddLink(int fromID, string fromName, int toID, string toName)
		{
			var link = new LinkMessageExt(FPatch, fromID, fromName, toID, toName);
			FPatch.DocumentElement.AppendChild(link);
			return link;
		}
		
		public BoundsMessageExt AddBounds()
		{
			var bounds = new BoundsMessageExt(FPatch, BoundsType.Window);
			FPatch.DocumentElement.AppendChild(bounds);
			return bounds;
		}
		
		override public string ToString()
		{
			return FPatch.OuterXml;
		}
	}
	
	public class NodeMessageExt: XmlElement
	{
		public NodeMessageExt(XmlDocument patch, int id): base(String.Empty, "NODE", String.Empty, patch)
		{
			ID = id;
		}
		
		public int ID
		{
			get {return int.Parse(GetAttribute("id"));}
			set {SetAttribute("id", value.ToString());}
		}
		
		public string SystemName
		{
			get {return GetAttribute("systemname");}
			set {SetAttribute("systemname", value);}
		}
		
		public ComponentMode ComponentMode
		{
			//get {return ComponentMode.Parse(GetAttribute("componentmode"));}
			set {SetAttribute("componentmode", value.ToString());}
		}
		
		public bool DeleteMe
		{
			get {return bool.Parse(GetAttribute("deleteme"));}
			set {SetAttribute("deleteme", value.ToString());}
		}
		
		public PinMessageExt AddPin(string pinName)
		{
			var pin = new PinMessageExt(OwnerDocument, pinName);
			AppendChild(pin);
			return pin;
		}
		
		public BoundsMessageExt AddBounds(BoundsType boundsType)
		{
			var bounds = new BoundsMessageExt(OwnerDocument, boundsType);
			AppendChild(bounds);
			return bounds;
		}
		
		override public string ToString()
		{
			return OuterXml;
		}
	}
	
	public class PinMessageExt: XmlElement
	{
		public PinMessageExt(XmlDocument patch, string pinName): base(String.Empty, "PIN", String.Empty, patch)
		{
			PinName = pinName;
		}
		
		public string PinName
		{
			get {return GetAttribute("pinname");}
			set {SetAttribute("pinname", value);}
		}
		
		public string Spread
		{
			get {return GetAttribute("values");}
			set {SetAttribute("values", value);}
		}
		
		override public string ToString()
		{
			return OuterXml;
		}
	}
	
	public class BoundsMessageExt: XmlElement
	{
		public BoundsMessageExt(XmlDocument patch, BoundsType boundsType): base(String.Empty, "BOUNDS", String.Empty, patch)
		{
			SetAttribute("type", boundsType.ToString());
		}
		
		private Rectangle FRectangle;
		public Rectangle Rectangle
		{
			get
			{
				return FRectangle;
			}
			
			set
			{
				FRectangle = value;
				SetAttribute("left", (value.Left*15).ToString());
				SetAttribute("top", (value.Top*15).ToString());
				SetAttribute("width", (value.Width*15).ToString());
				SetAttribute("height", (value.Height*15).ToString());
			}
		}
		
		private Rectangle FNodeRectangle;
		public Rectangle NodeRectangle
		{
			get
			{
				return FNodeRectangle;
			}
			
			set
			{
				FRectangle = value;
				SetAttribute("left", (value.Left).ToString());
				SetAttribute("top", (value.Top).ToString());
				SetAttribute("width", (value.Width).ToString());
				SetAttribute("height", (value.Height).ToString());
			}
		}
		
		override public string ToString()
		{
			return OuterXml;
		}
	}
	
	public class LinkMessageExt: XmlElement
	{
		public LinkMessageExt(XmlDocument patch, int fromID, string fromName, int toID, string toName): base(String.Empty, "LINK", String.Empty, patch)
		{
			FromID = fromID;
			FromName = fromName;
			ToID = toID;
			ToName = toName;
		}
		
		public int FromID
		{
			get {return int.Parse(GetAttribute("srcnodeid"));}
			set {SetAttribute("srcnodeid", value.ToString());}
		}
		
		public string FromName
		{
			get {return GetAttribute("srcpinname");}
			set {SetAttribute("srcpinname", value);}
		}
		
		public int ToID
		{
			get {return int.Parse(GetAttribute("dstnodeid"));}
			set {SetAttribute("dstnodeid", value.ToString());}
		}
		
		public string ToName
		{
			get {return GetAttribute("dstpinname");}
			set {SetAttribute("dstpinname", value);}
		}
		
		public string Linkstyle
		{
			get {return GetAttribute("linkstyle");}
			set {SetAttribute("linkstyle", value);}
		}
		public LinkPointExt AddLinkPoint(int x, int y)
		{
			var bounds = new LinkPointExt(OwnerDocument, x, y);
			AppendChild(bounds);
			return bounds;
		}
		
		override public string ToString()
		{
			return OuterXml;
		}
	}
	
	public class LinkPointExt: XmlElement
	{
		public LinkPointExt(XmlDocument patch, int x, int y): base(String.Empty, "LINKPOINT", String.Empty, patch)
		{
			X = x;
			Y = y;
		}
		public int X
		{
			get {return int.Parse(GetAttribute("x"));}
			set {SetAttribute("x", value.ToString());}
		}
		public int Y
		{
			get {return int.Parse(GetAttribute("y"));}
			set {SetAttribute("y", value.ToString());}
		}
		
	}
	#endregion VVVV.xmlMessage
}