
#region usings
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Xml;

using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.Graph;
using VVVV.PatchMessage;
using VVVV.Nodes;
#endregion usings


namespace VVVV.Snap
{
	public class SnapPatchsAndLinks
	{
		private IHDEHost FHDEHost;
		private int typ;
		
		public List<string> FOutXML = new List<string>();

			
		public  SnapPatchsAndLinks(IHDEHost Host)
		{
			FHDEHost = Host;
			typ = 0;
		}
		
		public void SnapNode(int delteLeft, int deltaRight,int deltaWidth,int deltaHeight)
		{
			FOutXML.Clear();
			string xmlSnippet = FHDEHost.GetXMLSnippetFromSelection();
			
			
			if ( xmlSnippet.Contains("<!DOCTYPE"))
			{
				
				xmlSnippet =  xmlSnippet.Remove(0,xmlSnippet.IndexOf(">")+1);
				XmlDocument selection = new XmlDocument();
				selection.LoadXml(xmlSnippet);
				XmlNodeList AllSelectedNodes = selection.SelectNodes("PATCH/NODE");
				
				
				foreach(XmlElement SingleNode in AllSelectedNodes)
				{
					int id = Convert.ToInt32(SingleNode.GetAttribute("id"));
					XmlNode SingleBound;
					
					PatchMessageExt PM = new PatchMessageExt();
					NodeMessageExt NODE = PM.AddNode(id);
					BoundsMessageExt BOUNDS;
					
					
					if ( SingleNode.GetAttribute("componentmode") == "InABox")
					{
						BOUNDS = NODE.AddBounds(BoundsType.Box);
						SingleBound = selection.SelectSingleNode("PATCH/NODE[@id='"+id.ToString()+"']/BOUNDS[@type='Box']");
					} else {
						BOUNDS = NODE.AddBounds(BoundsType.Node);
						SingleBound = selection.SelectSingleNode("PATCH/NODE[@id='"+id.ToString()+"']/BOUNDS[@type='Node']");
					}
					
					int left 	=	Math.Abs(Convert.ToInt32(SingleBound.Attributes["left"].Value)/delteLeft)*delteLeft;
					int top 	=	Math.Abs(Convert.ToInt32(SingleBound.Attributes["top"].Value)/deltaRight)*deltaRight;
					int width 	=	Math.Abs(Convert.ToInt32(SingleBound.Attributes["width"].Value)/deltaWidth)*deltaWidth;
					int height	=	Math.Abs(Convert.ToInt32(SingleBound.Attributes["height"].Value)/deltaHeight)*deltaHeight;
					BOUNDS.NodeRectangle = new Rectangle(left,top,width,height);
					
					if ( SingleNode.GetAttribute("systemname").Contains("IOBox"))
					{
						PinMessageExt PIN1 = NODE.AddPin("Rows");
						PIN1.Spread = (height/deltaHeight).ToString();
						PinMessageExt PIN2 = NODE.AddPin("Show Grid");
						PIN2.Spread = "1";
					}
					
					FHDEHost.SendXMLSnippet(FHDEHost.ActivePatchWindow.Node.NodeInfo.Filename, PM.ToString(), true);
					FOutXML.Add(PM.ToString());
				}
			}
		}	
	
		public void SnapLink(bool outputPins = true)
		{
			FOutXML.Clear();
			
			string xmlSnippet = FHDEHost.GetXMLSnippetFromSelection();
			List<INode2> allNODES = FHDEHost.ActivePatchWindow.Node.ToList();
			
			if ( xmlSnippet.Contains("<!DOCTYPE"))
			{
				xmlSnippet =  xmlSnippet.Remove(0,xmlSnippet.IndexOf(">")+1);
				XmlDocument selection = new XmlDocument();
				selection.LoadXml(xmlSnippet);
				XmlNodeList AllSelectedNodes = selection.SelectNodes("PATCH/NODE");
				
				
				foreach(XmlElement SingleNode in AllSelectedNodes)
				{
					int id = Convert.ToInt32(SingleNode.GetAttribute("id"));
					var Pins = (from node in allNODES where node.ID == id select node).First().Pins;
					
					List<XmlNode> AllConnectedLinks = new List<XmlNode>();
					List<int> AllConnectedLinksNR	= new List<int>();
					int PinNummer = 1;
					
					foreach(IPin2 Pin in Pins)
					{
						if (Pin.IsConnected())
						{
							XmlNodeList tmpLinks;
							if (outputPins)
							{
								tmpLinks = selection.SelectNodes("PATCH/LINK[@srcnodeid='"+id.ToString()+"'][@srcpinname='"+Pin.Name+"']");
							} else {
								tmpLinks = selection.SelectNodes("PATCH/LINK[@dstnodeid='"+id.ToString()+"'][@dstpinname='"+Pin.Name+"']");
							}
								
							if (tmpLinks.Count>0)
							{
								foreach(XmlNode LINK in tmpLinks)
								{
									AllConnectedLinks.Add(LINK);
									AllConnectedLinksNR.Add(PinNummer);
								}
								PinNummer++;
							}	
						}
					}
					
					
					for(int i = 0; i < AllConnectedLinks.Count; i++)
					{
						PatchMessageExt PM = new PatchMessageExt();
						LinkMessageExt LINK = PM.AddLink(Convert.ToInt32(AllConnectedLinks[i].Attributes["srcnodeid"].Value),AllConnectedLinks[i].Attributes["srcpinname"].Value,Convert.ToInt32(AllConnectedLinks[i].Attributes["dstnodeid"].Value),AllConnectedLinks[i].Attributes["dstpinname"].Value);
						
						LINK.Linkstyle = "VHV";
						
						int y;
						
						
						if (outputPins)
						{
							y = 60;
							if ( SingleNode.GetAttribute("componentmode") == "InABox")
							{
								y += Convert.ToInt32(selection.SelectSingleNode("PATCH/NODE[@id='"+id.ToString()+"']/BOUNDS[@type='Box']").Attributes["top"].Value);
								y += Convert.ToInt32(selection.SelectSingleNode("PATCH/NODE[@id='"+id.ToString()+"']/BOUNDS[@type='Box']").Attributes["height"].Value);
							} else {
								y += Convert.ToInt32(selection.SelectSingleNode("PATCH/NODE[@id='"+id.ToString()+"']/BOUNDS[@type='Node']").Attributes["top"].Value);
								y += 240;
								
							}
						} else {
							y = 0;
							if ( SingleNode.GetAttribute("componentmode") == "InABox")
							{
								y += Convert.ToInt32(selection.SelectSingleNode("PATCH/NODE[@id='"+id.ToString()+"']/BOUNDS[@type='Box']").Attributes["top"].Value);
							} else {
								y += Convert.ToInt32(selection.SelectSingleNode("PATCH/NODE[@id='"+id.ToString()+"']/BOUNDS[@type='Node']").Attributes["top"].Value);
							}
							y -= (AllConnectedLinks.Count*75)+150;
						}
						
						switch (typ)
						{
							case 1:
								y += 75*( PinNummer-AllConnectedLinksNR[i]);
								break;
							case 2:
								y += 75*( AllConnectedLinksNR[i]);
								break;
							case 0:
								if (i< AllConnectedLinks.Count/2)
								{
									y += 150*( AllConnectedLinksNR[i]);	
								}
								else
								{
									y += (150*( PinNummer-AllConnectedLinksNR[i]))-75;
								}
								break;
							default:
							break;
						}
						
						
						LinkPointExt Point1 = LINK.AddLinkPoint(0,y);
						LinkPointExt Point2 = LINK.AddLinkPoint(0,y);
						
						FHDEHost.SendXMLSnippet(FHDEHost.ActivePatchWindow.Node.NodeInfo.Filename, PM.ToString(), true);
						FOutXML.Add(PM.ToString());
					}
					
				}
				typ++;
				if (typ>2) typ=0;
			}
		}
		
	}
}