#region usings
using System;
using System.Collections.Generic;
using System.ComponentModel.Composition;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using System.Xml;
using System.Xml.Linq;

using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.Graph;
using VVVV.Core.Logging;
using VVVV.Utils.Streams;
using VVVV.PatchMessage;
using VVVV.Snap;
using VVVV.Nodes;
#endregion usings

namespace VVVV.Snap
{
	public class SnapWindows : IDisposable
	{
		#region field declaration
		private IHDEHost FHDEHost;
		//private PatchManagerNode FParent;
		
		public	IWindow2 		FActiveWindow;
		public  List<IWindow2>	FWindowAll 				= new List<IWindow2>();
		
		private  bool 			NodesSelected;
		private	List<INode2>	selectedNodes			= new List<INode2>();
		private List<int>		AllScreensWorkingArea	= new List<int>();
		private List<int>		horizontalSnaps			= new List<int>();
		private List<int>		verticalSnaps			= new List<int>();
		
		public List<string> FOutXML				= new List<string>();
		public List<string> FOutFile			= new List<string>();
		public List<int> FOutHorizontalSnaps 	= new List<int>();
		public List<int> FOutVerticalSnaps 		= new List<int>();
		
		// NodesCollection
		public INode2 inspektor;
		public string inspektorID;
		public INode2 nodeBrowser;
		public INode2 finder;
		public INode2 projektexplorer;
		
		#endregion field declaration
		
		#region constructor/destructor
		public  SnapWindows(IHDEHost Host, PatchManagerNode Parent)
		{
			FHDEHost = Host;
			FHDEHost.WindowAdded 				+= FHDEHost_WindowAdded;
			FHDEHost.WindowRemoved 				+= FHDEHost_WindowRemoved;
			FHDEHost.WindowSelectionChanged 	+= FHDEHost_WindowSelectionChanged;
			FHDEHost.NodeSelectionChanged 		+= FHDEHost_NodeSelectionEventHandler;
			FActiveWindow 						=  FHDEHost.ActivePatchWindow;
		}
		
		public void Dispose()
		{
			FHDEHost.WindowAdded 				-= FHDEHost_WindowAdded;
			FHDEHost.WindowRemoved 				-= FHDEHost_WindowRemoved;
			FHDEHost.WindowSelectionChanged 	-= FHDEHost_WindowSelectionChanged;
			FHDEHost.NodeSelectionChanged 		-= FHDEHost_NodeSelectionEventHandler;
		}
		#endregion constructor/destructor
		
		#region Show & Hide
		public void Hide()
		{
			if (NodesSelected)
			{
				foreach(INode2 node in selectedNodes)
				{
					HideNode(node);
				}
			}
			else
			{
				HideWindow();
			}
		}
		
		public void ShowInspektor()
		{
			if (inspektor.Window != null && inspektor.Window.IsVisible)
			{
				HideWindow(inspektor.Window);
				return;
			}
			string xmlSnipplet = "<PATCH><NODE id=\"" + inspektor.ID.ToString() + "\"  createme=\"pronto\"></PATCH>"; 
			string FileName = inspektor.Parent.NodeInfo.Filename;
			FHDEHost.SendXMLSnippet(FileName,xmlSnipplet, false);
			
			ShowNode(inspektor);
			//FHDEHost.ShowGUI(inspektor);				

		}
		public void ShowNodeBrowser()
		{
			if (nodeBrowser.Window != null && nodeBrowser.Window.IsVisible)
			{
				HideWindow(nodeBrowser.Window);
				return;
			}
			ShowNode(nodeBrowser);
		}
		public void ShowFinder()
		{
			if (finder.Window != null && finder.Window.IsVisible)
			{
				HideWindow(finder.Window);
				return;
			}
			ShowNode(finder);
		}
		public void ShowProjektexplorer()
		{
			if (projektexplorer.Window != null && projektexplorer.Window.IsVisible)
			{
				HideWindow(projektexplorer.Window);
				return;
			}
			ShowNode(projektexplorer);
		}
		
		private void HideWindow()
		{
			HideWindow(FActiveWindow);
		}
		
		private void HideWindow(IWindow2 selectedWindow)
		{
			INode2 selectedNode = selectedWindow.Node;
			
			if (selectedNode.Parent == FHDEHost.RootNode )
			{
				HideNode(selectedNode);
			}
			else if (selectedNode.Parent.Window != null)
			{
				HideNode(selectedNode);
			} else {
				ShowNode(selectedNode.Parent);
				HideNode(selectedNode);
				HideNode(selectedNode.Parent);
			}
		}
		
		private void HideNode(INode2 selectedNode)
		{
			FHDEHost.SetComponentMode(selectedNode, ComponentMode.InABox);
			FHDEHost.SetComponentMode(selectedNode, ComponentMode.Hidden);
		}
		
		private void ShowNode(INode2 selectedNode)
		{
			FHDEHost.SetComponentMode(selectedNode , ComponentMode.InAWindow);
		}
		#endregion Show & Hide
		
		#region SetWindow
		public void SetNewWindowPosition(Rectangle NewWindowPosition)
		{
			INode2 selectedNode = FActiveWindow.Node;
			
			HideWindow(FActiveWindow);
			
			string FileName = selectedNode.Parent.NodeInfo.Filename;
			PatchMessageExt pm = new PatchMessageExt();
			var _node1 = pm.AddNode(selectedNode.ID);
			_node1.ComponentMode = ComponentMode.InAWindow;
			var _bounds1 = _node1.AddBounds(BoundsType.Window);
			_bounds1.Rectangle = NewWindowPosition;
			
			FHDEHost.SendXMLSnippet(FileName, pm.ToString(), false);
			
			FOutXML.Clear();
			FOutFile.Clear();
			FOutXML.Add(pm.ToString());
			FOutFile.Add(FileName);
		}
		#endregion SetNewWindowPosition
		
		#region SnapWindows
		public void SnapActiveWindow(int delta)
		{
			CalcAllSnaps();
			SetNewWindowPosition(CalcNewSnapWindowPosition(delta));
		}
		
		public void goLeft()
		{
			CalcAllSnaps();
			Rectangle myPlace	= FActiveWindow.Node.Window.Bounds;
			int indexLeft		= horizontalSnaps.IndexOf(myPlace.Left);
			int indexRight		= horizontalSnaps.IndexOf(myPlace.Right);
			
			if (indexLeft > 0 && (indexRight-indexLeft) == 1 )
			{
				myPlace = new Rectangle(horizontalSnaps[indexLeft-1],myPlace.Top,horizontalSnaps[indexLeft]-horizontalSnaps[indexLeft-1],myPlace.Height);
				SetNewWindowPosition(myPlace);
			}
			else if ((indexRight-indexLeft) > 1)
			{
				myPlace = new Rectangle(horizontalSnaps[indexLeft],myPlace.Top,horizontalSnaps[indexRight-1]-horizontalSnaps[indexLeft],myPlace.Height);
				SetNewWindowPosition(myPlace);
			}
		}
		
		public void crowLeft()
		{
			CalcAllSnaps();
			Rectangle myPlace = FActiveWindow.Node.Window.Bounds;
			int index =horizontalSnaps.IndexOf(myPlace.Left);
			
			if (index > 0)
			{
				myPlace = new Rectangle(horizontalSnaps[index-1],myPlace.Top,horizontalSnaps[index]-horizontalSnaps[index-1]+myPlace.Width,myPlace.Height);
				SetNewWindowPosition(myPlace);
			}
		}
		
		public void goRight()
		{
			CalcAllSnaps();
			Rectangle myPlace	= FActiveWindow.Node.Window.Bounds;
			int indexLeft		= horizontalSnaps.IndexOf(myPlace.Left);
			int indexRight		=horizontalSnaps.IndexOf(myPlace.Right);
			
			if (indexRight < (horizontalSnaps.Count-1) && (indexRight-indexLeft) == 1 )
			{
				myPlace = new Rectangle(horizontalSnaps[indexRight],myPlace.Top,horizontalSnaps[indexRight+1]-horizontalSnaps[indexRight],myPlace.Height);
				SetNewWindowPosition(myPlace);
			}
			else if ((indexRight-indexLeft) > 1)
			{
				myPlace = new Rectangle(horizontalSnaps[indexLeft+1],myPlace.Top,horizontalSnaps[indexRight]-horizontalSnaps[indexLeft+1],myPlace.Height);
				SetNewWindowPosition(myPlace);
				
			}
		}
		
		public void crowRight()
		{
			CalcAllSnaps();
			Rectangle myPlace = FActiveWindow.Node.Window.Bounds;
			int index =horizontalSnaps.IndexOf(myPlace.Right);
			
			if (index < (horizontalSnaps.Count-1))
			{
				myPlace = new Rectangle(myPlace.Left,myPlace.Top,horizontalSnaps[index+1]-horizontalSnaps[index]+myPlace.Width,myPlace.Height);
				SetNewWindowPosition(myPlace);
			}
		}
		
		public void goUp()
		{
			CalcAllSnaps();
			Rectangle myPlace	= FActiveWindow.Node.Window.Bounds;
			int indexTop		= verticalSnaps.IndexOf(myPlace.Top);
			int indexBottom		= verticalSnaps.IndexOf(myPlace.Bottom);
			
			if (indexTop > 0 && (indexBottom-indexTop) == 1)
			{
				myPlace = new Rectangle(myPlace.Left,verticalSnaps[indexTop-1],myPlace.Width,verticalSnaps[indexTop]-verticalSnaps[indexTop-1]);
				SetNewWindowPosition(myPlace);
			}
			else if ((indexBottom-indexTop) > 1)
			{
				myPlace = new Rectangle(myPlace.Left,verticalSnaps[indexTop],myPlace.Width,verticalSnaps[indexBottom-1]-verticalSnaps[indexTop]);
				SetNewWindowPosition(myPlace);
				
			}
		}
		
		public void crowUp()
		{
			CalcAllSnaps();
			Rectangle myPlace	= FActiveWindow.Node.Window.Bounds;
			int indexTop		= verticalSnaps.IndexOf(myPlace.Top);
			
			if (indexTop > 0)
			{
				myPlace = new Rectangle(myPlace.Left,verticalSnaps[indexTop-1],myPlace.Width,verticalSnaps[indexTop]-verticalSnaps[indexTop-1]+myPlace.Height);
				SetNewWindowPosition(myPlace);
			}
		}
		
		public void goDown()
		{
			CalcAllSnaps();
			Rectangle myPlace	= FActiveWindow.Node.Window.Bounds;
			int indexTop		= verticalSnaps.IndexOf(myPlace.Top);
			int indexBottom		=verticalSnaps.IndexOf(myPlace.Bottom);
			
			if (indexBottom < (verticalSnaps.Count-1) && (indexBottom-indexTop) == 1)
			{
				myPlace = new Rectangle(myPlace.Left,verticalSnaps[indexBottom],myPlace.Width,verticalSnaps[indexBottom+1]-verticalSnaps[indexBottom]);
				SetNewWindowPosition(myPlace);
			}
			else if ((indexBottom-indexTop) > 1)
			{
				myPlace = new Rectangle(myPlace.Left,verticalSnaps[indexTop+1],myPlace.Width,verticalSnaps[indexBottom]-verticalSnaps[indexTop+1]);
				SetNewWindowPosition(myPlace);
			}
		}
		
		public void crowDown()
		{
			CalcAllSnaps();
			Rectangle myPlace	= FActiveWindow.Node.Window.Bounds;
			int indexBottom		=verticalSnaps.IndexOf(myPlace.Bottom);
			
			if (indexBottom < (verticalSnaps.Count-1))
			{
				myPlace = new Rectangle(myPlace.Left,myPlace.Top,myPlace.Width,verticalSnaps[indexBottom+1]-verticalSnaps[indexBottom]+myPlace.Height);
				SetNewWindowPosition(myPlace);
			}
		}
		#endregion SnapWindows
		
		#region CalcSnaps and NewPos
		private void CalcAllSnaps()
		{
			horizontalSnaps.Clear();
			verticalSnaps.Clear();
			
			GetAllScreensWorkingArea(true);
			
			
			foreach (IWindow2 window in FWindowAll)
			{
				if ( window.IsVisible)
				{
					Rectangle myPlace = window.Bounds;
					if (horizontalSnaps.IndexOf(myPlace.Left) < 0) horizontalSnaps.Add(myPlace.Left);
					if (horizontalSnaps.IndexOf(myPlace.Right) < 0) horizontalSnaps.Add(myPlace.Right);
					if (verticalSnaps.IndexOf(myPlace.Top) < 0) verticalSnaps.Add(myPlace.Top);
					if (verticalSnaps.IndexOf(myPlace.Bottom) < 0) verticalSnaps.Add(myPlace.Bottom);
				}
			}
			
			
			horizontalSnaps.Sort();
			FOutHorizontalSnaps.Clear();
			for( int i = 0 ; i <  horizontalSnaps.Count; i++)
			{
				FOutHorizontalSnaps.Add(horizontalSnaps[i]);
			}
			
			verticalSnaps.Sort();
			FOutVerticalSnaps.Clear();
			for( int i = 0 ; i <  verticalSnaps.Count; i++)
			{
				FOutVerticalSnaps.Add(verticalSnaps[i]);
			}
			
		}
		
		private Rectangle CalcNewSnapWindowPosition(int delta)
		{
			int deltaMax = delta;
			Rectangle myPlace = FActiveWindow.Bounds;
			
			List<int> deltaLeft = new List<int>();
			List<int> deltaTop = new List<int>();
			List<int> deltaRight = new List<int>();
			List<int> deltaBottom = new List<int>();
			
			foreach(int snappoint in horizontalSnaps)
			{
				if (Math.Abs(snappoint - myPlace.Left) < deltaMax) 		deltaLeft.Add(snappoint - myPlace.Left);
				if (Math.Abs(snappoint - myPlace.Right) < deltaMax)		deltaRight.Add(snappoint - myPlace.Right);
			}
			foreach(int snappoint in verticalSnaps)
			{
				if (Math.Abs(snappoint - myPlace.Top) < deltaMax)		deltaTop.Add(snappoint - myPlace.Top);
				if (Math.Abs(snappoint - myPlace.Bottom) < deltaMax)	deltaBottom.Add(snappoint - myPlace.Bottom);
			}
			
			deltaLeft.Sort(CompareDeltaToZero);
			deltaTop.Sort(CompareDeltaToZero);
			deltaRight.Sort(CompareDeltaToZero);
			deltaBottom.Sort(CompareDeltaToZero);
			
			if (deltaLeft.Count>1) deltaLeft.RemoveAt(0);
			if (deltaTop.Count>1) deltaTop.RemoveAt(0);
			if (deltaRight.Count>1) deltaRight.RemoveAt(0);
			if (deltaBottom.Count>1) deltaBottom.RemoveAt(0);
			
			int left	= (myPlace.Left+deltaLeft[0]);
			int top		= (myPlace.Top+deltaTop[0]);
			int right	= (myPlace.Right+deltaRight[0]);
			int bottom	= (myPlace.Bottom+deltaBottom[0]);
			
			return new Rectangle(left,top,right-left,bottom-top);
		}
		
		private static int CompareDeltaToZero(int x, int y)
		{
			if (Math.Abs(x) >  Math.Abs(y)) return 1;
			if (Math.Abs(x) == Math.Abs(y)) return 0;
			if (Math.Abs(x) <  Math.Abs(y)) return -1;
			return 0;
		}
		
		private void GetAllScreensWorkingArea(bool setSnaps)
		{
			
			AllScreensWorkingArea	= new List<int>();
			foreach( Screen screen in Screen.AllScreens)
			{
				AllScreensWorkingArea.Add(screen.WorkingArea.Left);
				AllScreensWorkingArea.Add(screen.WorkingArea.Top);
				AllScreensWorkingArea.Add(screen.WorkingArea.Right);
				AllScreensWorkingArea.Add(screen.WorkingArea.Bottom);
				if (setSnaps)
				{
					if (horizontalSnaps.IndexOf(screen.WorkingArea.Left) < 0) horizontalSnaps.Add(screen.WorkingArea.Left);
					if (horizontalSnaps.IndexOf(screen.WorkingArea.Right) < 0) horizontalSnaps.Add(screen.WorkingArea.Right);
					if (verticalSnaps.IndexOf(screen.WorkingArea.Top) < 0) verticalSnaps.Add(screen.WorkingArea.Top);
					if (verticalSnaps.IndexOf(screen.WorkingArea.Bottom) < 0) verticalSnaps.Add(screen.WorkingArea.Bottom);
				}
			}
		}
		#endregion CalcSnaps and NewPos
		
		#region WindowHandeling
		
		private void FHDEHost_WindowAdded(object sender, WindowEventArgs args)
		{
			if (args.Window.Caption != "Kommunikator")
			{
				FWindowAll.Add(args.Window);
			} 
			
			/*if (tmp.Name == "Inspektor (VVVV)")
			{
				//inspektor = args.Window.Node;
			}
			else if (args.Window.Node.NodeInfo.Systemname == "NodeBrowser (VVVV)")
			{
				nodeBrowser = args.Window.Node;
			}
			else if (args.Window.Node.NodeInfo.Systemname == "Finder (HDE)")
			{
				finder = args.Window.Node;
			}
			else if (args.Window.Node.NodeInfo.Systemname == "ProjectExplorer (VVVV)")
			{
				projektexplorer = args.Window.Node;
			}*/
			selectedNodes.Clear();
			NodesSelected = false;
		}
		
		private void FHDEHost_WindowRemoved(object sender, WindowEventArgs args)
		{
			FWindowAll.Remove(args.Window);
			
			selectedNodes.Clear();
			NodesSelected = false;
		}
		
		private void FHDEHost_WindowSelectionChanged(object sender, WindowEventArgs args)
		{
			FActiveWindow = args.Window;
			FWindowAll.Remove(args.Window);
			FWindowAll.Insert(0,args.Window);
			selectedNodes.Clear();
			NodesSelected = false;
		}
		
		
		private void FHDEHost_NodeSelectionEventHandler(object sender, NodeSelectionEventArgs args)
		{
			selectedNodes.Clear();
			NodesSelected = false;
			if (args.Nodes.Length > 0)
			{
				NodesSelected = true;
				
				foreach(INode2 node in args.Nodes)
				{
					selectedNodes.Add(node);
				}
			}
		}
		
		#endregion WindowHandeling
		
	}
	
}